from .utils.belief_certainties_calculator_util import calculate_belief_certainties
from typing import Optional
import uuid
from datetime import datetime
from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.utils.entity_formatters import (
    format_belief_str,
    format_belief_repr,
)


class Belief:
    def __init__(
        self,
        agent_id: str,
        id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
    ):
        self.id = id if id else str(uuid.uuid4())
        self.agent_id = agent_id
        self._statements = []
        self._modifier = None
        self._probability = None
        self._validity = None
        if created_at is None:
            created_at = datetime.now()
        if updated_at is None:
            updated_at = created_at
        self.created_at = created_at
        self.updated_at = updated_at

    @property
    def subject_notion(self):
        if not self._statements:
            raise ValueError("Belief has no statements")
        return self._statements[0].subject_notion

    @property
    def complement_notion(self):
        if not self._statements:
            raise ValueError("Belief has no statements")
        return self._statements[0].complement_notion

    @property
    def relation(self):
        if not self._statements:
            raise ValueError("Belief has no statements")
        return self._statements[0].relation
    
    
    def add_statement(self, statement: StatementBase):
        if any(s.id == statement.id for s in self._statements if s.id is not None):
            app_logger.debug(f"Skipping duplicate statement in belief (statement_id: {statement.id}, belief_id: {self.id})")
            return
        
        statement_type = type(statement).__name__
        app_logger.debug(f"Adding statement to belief: type={statement_type} (statement_id: {statement.id}, belief_id: {self.id})")
        
        if not self._statements:
            self._modifier = statement.modifier
        
        self._statements.append(statement)
        self._recompute_certainties()
        self.updated_at = datetime.now()
        app_logger.debug(f"Added statement to belief, total statements: {len(self._statements)} (statement_id: {statement.id}, belief_id: {self.id})")

    @property
    def statements(self):
        return list(self._statements)
    
    @property
    def unfiltered_statements(self):
        return list(self._statements)
    
    def _calculate_certainties(self, statements) -> list:
        if not statements:
            raise ValueError("No statements to calculate certainties from")
        return [(s.validity, s.probability) for s in statements]

    def _recompute_certainties(self) -> None:
        if not self._statements:
            self._probability = None
            self._validity = None
            return
        certainties = self._calculate_certainties(self.statements)
        probability, validity = calculate_belief_certainties(certainties)
        self._probability = probability
        self._validity = validity

    @property
    def probability(self) -> float:
        if self._probability is None:
            self._recompute_certainties()
        if self._probability is None:
            raise ValueError("Trying to calculate probability for belief without statements")
        return self._probability

    @property
    def validity(self) -> float:
        if self._validity is None:
            self._recompute_certainties()
        if self._validity is None:
            raise ValueError("Trying to calculate validity for belief without statements")
        return self._validity

    @property
    def modifier(self) -> Optional[Modifier]:
        return self._modifier
    
    @property
    def location(self) -> Optional[str]:
        return self._modifier.location if self._modifier else None

    @property
    def from_time(self):
        return self._modifier.from_time if self._modifier else None

    @property
    def to_time(self):
        return self._modifier.to_time if self._modifier else None

    @property
    def deontic_modality(self):
        return self._modifier.deontic_modality if self._modifier else None

    @property
    def quantity(self) -> Optional[float]:
        return self._modifier.quantity if self._modifier else None

    def __str__(self) -> str:
        return format_belief_str(self)

    def __repr__(self) -> str:
        return format_belief_repr(self)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Belief):
            return False
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)


from .enums import AuthorizationOperation
from .services import AuthorizationService
from .guards import AuthorizationGuard, AuthorizationGuardError, AuthorizationErrorReason

__all__ = [
    "AuthorizationOperation",
    "AuthorizationService",
    "AuthorizationGuard",
    "AuthorizationGuardError",
    "AuthorizationErrorReason",
]




from typing import Dict
from analogai.foundation.modules.user.user import User


class UserStorageMapper:
    @staticmethod
    def to_dict(user: User) -> Dict:
        result = {
            "id": user.id,
            "name": user.name or "",
            "email": user.email or ""
        }
        if user.timezone is not None:
            result["timezone"] = user.timezone
        if user.nationality is not None:
            result["nationality"] = user.nationality
        if user.location is not None:
            result["location"] = user.location
        return result

    @staticmethod
    def from_dict(data: Dict) -> User:
        user = User(
            id=data.get("id"),
            name=data.get("name", ""),
            email=data.get("email", ""),
            timezone=data.get("timezone"),
            nationality=data.get("nationality"),
            location=data.get("location")
        )
        return user




from __future__ import annotations

from analogai.foundation.settings import config
from analogai.foundation.infrastructure.embeddings.base import EmbeddingProvider
from analogai.foundation.infrastructure.embeddings.local_provider import LocalSentenceTransformerProvider
from analogai.foundation.infrastructure.embeddings.openai_provider import OpenAIEmbeddingProvider
from analogai.foundation.infrastructure.embeddings.azure_openai_provider import AzureOpenAIEmbeddingProvider
from analogai.foundation.infrastructure.embeddings.azure_foundry_provider import AzureFoundryEmbeddingProvider


DEFAULT_LOCAL_MODEL = "all-MiniLM-L6-v2"


def create_embedding_provider() -> EmbeddingProvider:
    embedding_type = str(config.get("embed.type", "local")).strip().lower()
    embedding_provider = str(config.get("embed.provider", "")).strip().lower()
    embedding_model = config.get("embed.model", "") or ""
    azure_openai_endpoint = config.get("AZURE_OPENAI_ENDPOINT", "") or ""
    azure_openai_api_key = config.get("AZURE_OPENAI_API_KEY", "") or ""
    azure_foundry_endpoint = config.get("AZURE_AI_FOUNDRY_ENDPOINT", "") or ""
    azure_foundry_api_key = config.get("AZURE_AI_FOUNDRY_API_KEY", "") or ""
    openai_api_key = config.get("OPENAI_API_KEY", "") or ""
    azure_openai_version = config.get("azure.openai.version", "2024-02-15-preview")

    if embedding_type == "local":
        model_name = embedding_model or DEFAULT_LOCAL_MODEL
        return LocalSentenceTransformerProvider(model_name=model_name)

    if embedding_provider:
        provider = embedding_provider
        if provider == "azure":
            provider = "azure_openai" if azure_openai_endpoint else "azure_foundry"
        if provider == "azure_openai":
            return AzureOpenAIEmbeddingProvider(
                endpoint=azure_openai_endpoint,
                api_key=azure_openai_api_key,
                deployment_name=embedding_model,
                api_version=azure_openai_version,
            )
        if provider == "azure_foundry":
            return AzureFoundryEmbeddingProvider(
                endpoint=azure_foundry_endpoint,
                api_key=azure_foundry_api_key,
                deployment_name=embedding_model,
                api_version=azure_openai_version,
            )
        if provider == "openai":
            return OpenAIEmbeddingProvider(
                api_key=openai_api_key,
                model_name=embedding_model,
            )
        raise ValueError(
            "Unknown EMBEDDING_PROVIDER value. Use one of: local, openai, azure_openai, azure_foundry."
        )

    if azure_openai_endpoint:
        return AzureOpenAIEmbeddingProvider(
            endpoint=azure_openai_endpoint,
            api_key=azure_openai_api_key,
            deployment_name=embedding_model,
            api_version=azure_openai_version,
        )

    if azure_foundry_endpoint:
        return AzureFoundryEmbeddingProvider(
            endpoint=azure_foundry_endpoint,
            api_key=azure_foundry_api_key,
            deployment_name=embedding_model,
            api_version=azure_openai_version,
        )

    if openai_api_key:
        return OpenAIEmbeddingProvider(
            api_key=openai_api_key,
            model_name=embedding_model,
        )

    raise ValueError(
        "EMBEDDING_TYPE is set to 'api' but no embedding provider credentials were configured."
    )



import unittest
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.common.utils.modifier_utils import ModifierBuilder
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.enums.deontic_modality import DeonticModality

class TestModifierBuilder(unittest.TestCase):
    
    def test_build_empty_context(self):
        builder = ModifierBuilder()
        context = builder.build()
        
        self.assertIsNone(context.location)
        self.assertIsNone(context.deontic_modality)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_with_location(self):
        context = ModifierBuilder().with_location("Paris").build()
        
        self.assertEqual(context.location, "Paris")
        self.assertIsNone(context.deontic_modality)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_with_deontic_modality(self):
        context = ModifierBuilder().with_deontic_modality(DeonticModality.FORBIDDEN).build()
        
        self.assertIsNone(context.location)
        self.assertEqual(context.deontic_modality, DeonticModality.FORBIDDEN)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_with_from_time_full(self):
        context = ModifierBuilder().with_from_time(
            year=2024, month=1, day=1, hour=10, minute=30
        ).build()
        
        self.assertIsNotNone(context.from_time)
        self.assertEqual(context.from_time.year, 2024)
        self.assertEqual(context.from_time.month, 1)
        self.assertEqual(context.from_time.day, 1)
        self.assertEqual(context.from_time.hour, 10)
        self.assertEqual(context.from_time.minute, 30)
    
    def test_with_from_time_partial(self):
        context = ModifierBuilder().with_from_time(year=2024, month=1).build()
        
        self.assertIsNotNone(context.from_time)
        self.assertEqual(context.from_time.year, 2024)
        self.assertEqual(context.from_time.month, 1)
        self.assertIsNone(context.from_time.day)
        self.assertIsNone(context.from_time.hour)
        self.assertIsNone(context.from_time.minute)
    
    def test_with_to_time_full(self):
        context = ModifierBuilder().with_to_time(
            year=2024, month=12, day=31, hour=23, minute=59
        ).build()
        
        self.assertIsNotNone(context.to_time)
        self.assertEqual(context.to_time.year, 2024)
        self.assertEqual(context.to_time.month, 12)
        self.assertEqual(context.to_time.day, 31)
        self.assertEqual(context.to_time.hour, 23)
        self.assertEqual(context.to_time.minute, 59)
    
    def test_fluent_interface(self):
        context = (ModifierBuilder()
                   .with_location("Tokyo")
                   .with_deontic_modality(DeonticModality.PERMITTED)
                   .with_from_time(year=2024, month=1)
                   .with_to_time(year=2024, month=12)
                   .build())
        
        self.assertEqual(context.location, "Tokyo")
        self.assertEqual(context.deontic_modality, DeonticModality.PERMITTED)
        self.assertIsNotNone(context.from_time)
        self.assertEqual(context.from_time.year, 2024)
        self.assertEqual(context.from_time.month, 1)
        self.assertIsNotNone(context.to_time)
        self.assertEqual(context.to_time.year, 2024)
        self.assertEqual(context.to_time.month, 12)
    
    def test_multiple_calls_override(self):
        builder = ModifierBuilder()
        context1 = builder.with_location("Paris").build()
        context2 = builder.with_location("London").build()
        
        self.assertEqual(context1.location, "Paris")
        self.assertEqual(context2.location, "London")

if __name__ == '__main__':
    unittest.main()



import spacy

from analogai.foundation.settings import config


_nlp = None

_PLACEHOLDER_PREFIXES = ("respondent:", "agent:", "company:")
_RESTRICTED_WORDS = {
    "birth",
    "lain",
    "laylie",
    "lie",
}


def _is_placeholder_identifier(text: str) -> bool:
    if not text or not text.strip():
        return False
    lowered = text.strip().lower()
    return any(lowered.startswith(prefix) for prefix in _PLACEHOLDER_PREFIXES)


def _get_nlp():
    global _nlp
    if _nlp is None:
        _nlp = spacy.load("en_core_web_sm")
    return _nlp


def _is_restricted_word(text: str) -> bool:
    if not text or not text.strip():
        return False
    return text.strip().lower() in _RESTRICTED_WORDS


def _is_named_entity_token(token) -> bool:
    return token.ent_iob_ != "O" or bool(token.ent_type_)


def _contains_named_entity(doc) -> bool:
    if not doc:
        return False
    return any(_is_named_entity_token(token) for token in doc)


def _contains_restricted_word(doc) -> bool:
    if not doc:
        return False
    return any(_is_restricted_word(token.text) for token in doc)


def is_likely_proper_noun(token) -> bool:
    if token.pos_ == "PROPN":
        return True
    
    text = token.text
    if text and text[0].isupper() and len(text) > 1:
        return True
    
    return False


def lemmatize_text(text: str, preserve_proper: bool = True) -> str:
    if not text or not text.strip():
        return ""
    if _is_placeholder_identifier(text):
        return text.strip()
    if not config.get("lemmatize", False):
        return text.strip()
    
    nlp = _get_nlp()
    doc = nlp(text.strip())
    result_tokens = []
    
    for token in doc:
        if token.is_punct or token.is_space:
            continue

        if token.text.lower() == "born":
            lemma = "birth"
        elif _is_named_entity_token(token) or _is_restricted_word(token.text):
            lemma = token.text
        elif preserve_proper and is_likely_proper_noun(token):
            lemma = token.text
        else:
            lemma = token.lemma_.lower()
        
        result_tokens.append(lemma)
    
    return " ".join(result_tokens)


if __name__ == "__main__":
    test_cases = [
        "Socrates",
        "iPhone",
        "John",
        "New York",
        "apples",
        "humans",
        "philosophers",
        "born humans",
        "George bought an iPhone",
        "Giorgi was born at 1991",
        "born",
    ]

    for text in test_cases:
        lemmatized = lemmatize_text(text)
        print(f"Original: '{text}' -> Lemmatized: '{lemmatized}'")


import unittest
from unittest import TestCase
from unittest.mock import Mock
from datetime import datetime, timedelta
from time import sleep
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement as Statement
from analogai.foundation.core.value_objects.relation import Relation

from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.agent.agent import Agent
from analogai.foundation.infrastructure.repositories.in_memory.in_memory_belief_repository_impl import InMemoryBeliefRepositoryImpl
from analogai.foundation.infrastructure.storage.storage_adapter import StorageAdapter
from analogai.foundation.infrastructure.mappers import BeliefStorageMapper, StatementStorageMapper
from copy import copy
from analogai.foundation.settings import config

class TestInMemoryBeliefRepositoryImpl(TestCase):
    def setUp(self):
        self._config_override = config.override(**{"vector.enabled": False})
        self._config_override.__enter__()
        self.storage_adapter = Mock(spec=StorageAdapter)
        self.repository = InMemoryBeliefRepositoryImpl(storage_adapter=self.storage_adapter)
        self.user = User(id="user_id", name="Test User", email="test@example.com")
        self.agent = Agent(id="agent_id", knowledge_base="Test KB", creator=self.user, roles=[])
        self.subject_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="human", id="human_id")
        self.complement_notion = Notion(user_id=self.user.id, agent_id=self.agent.id, caption="mortal", id="mortal_id")
        self.statement = Statement(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=Relation.from_string("is"),
            complement_notion=self.complement_notion,
            probability=0.8,
            validity=0.9,
            id="stmt_id"
        )
        self.belief = Belief(agent_id=self.agent.id, id="belief_id")
        self.belief.add_statement(self.statement)

    def test_find(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        result = self.repository.find(self.subject_notion, self.complement_notion, Relation.from_string("is"))
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id",
            'relation': "is"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(result.id, self.belief.id)

    def test_find_no_result(self):
        self.storage_adapter.retrieve_by_attributes.return_value = []
        
        result = self.repository.find(self.subject_notion, self.complement_notion, Relation.from_string("is"))
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id",
            'relation': "is"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertIsNone(result)

    def test_find_by_id(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        
        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "beliefs":
                return belief_data
            elif collection == "statements":
                return statement_data
            return None
        
        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect
        
        result = self.repository.find_by_id("belief_id")
        
        self.assertEqual(self.storage_adapter.retrieve_by_id.call_count, 2)
        self.storage_adapter.retrieve_by_id.assert_any_call("beliefs", "belief_id")
        self.storage_adapter.retrieve_by_id.assert_any_call("statements", "stmt_id")
        self.assertEqual(result.id, self.belief.id)

    def test_find_by_id_not_found(self):
        self.storage_adapter.retrieve_by_id.return_value = None
        
        result = self.repository.find_by_id("nonexistent_id")
        
        self.storage_adapter.retrieve_by_id.assert_called_once_with("beliefs", "nonexistent_id")
        self.assertIsNone(result)

    def test_delete(self):
        self.repository.delete(self.belief)
        
        self.storage_adapter.delete.assert_called_once_with("beliefs", "belief_id")

    def test_search_by_notions(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_notions(self.subject_notion, self.complement_notion)
        
        expected_query = {
            'subject_notion.id': "human_id",
            'complement_notion.id': "mortal_id"
        }
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_search_by_subject_notion(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_subject_notion(self.subject_notion)
        
        expected_query = {'subject_notion.id': "human_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_search_by_complement_notion(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.search_by_complement_notion(self.complement_notion)
        
        expected_query = {'complement_notion.id': "mortal_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_find_by_agent_id(self):
        belief_data = BeliefStorageMapper.to_dict(self.belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        results = self.repository.find_by_agent_id("agent_id")
        
        expected_query = {'agent_id': "agent_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, self.belief.id)

    def test_find_by_agent_id_no_results(self):
        self.storage_adapter.retrieve_by_attributes.return_value = []
        
        results = self.repository.find_by_agent_id("nonexistent_agent_id")
        
        expected_query = {'agent_id': "nonexistent_agent_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(results, [])

    def test_belief_has_created_at_timestamp(self):
        self.assertTrue(hasattr(self.belief, 'created_at'))
        self.assertIsInstance(self.belief.created_at, datetime)

    def test_belief_has_updated_at_timestamp(self):
        self.assertTrue(hasattr(self.belief, 'updated_at'))
        self.assertIsInstance(self.belief.updated_at, datetime)

    def test_belief_timestamps_are_set_on_creation(self):
        self.assertIsNotNone(self.belief.created_at)
        self.assertIsNotNone(self.belief.updated_at)
        self.assertEqual(self.belief.created_at, self.belief.updated_at)

    def test_belief_updated_at_changes_when_statement_added(self):
        created_at = self.belief.created_at
        updated_at = self.belief.updated_at
        
        new_statement = Statement(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=Relation.from_string("is"),
            complement_notion=self.complement_notion,
            probability=0.7,
            validity=0.8,
            id="stmt_id_2"
        )
        
        sleep(0.001)
        self.belief.add_statement(new_statement)
        
        self.assertEqual(self.belief.created_at, created_at)
        self.assertGreater(self.belief.updated_at, updated_at)

    def test_repository_stores_belief_with_timestamps(self):
        belief = Belief(agent_id=self.agent.id)
        belief.add_statement(self.statement)
        
        self.repository.save(belief)
        
        self.assertEqual(self.storage_adapter.store.call_count, 1)
        stored_data = self.storage_adapter.store.call_args[0][1]
        self.assertIn('created_at', stored_data)
        self.assertIn('updated_at', stored_data)
        self.assertIsNotNone(stored_data.get('created_at'))
        self.assertIsNotNone(stored_data.get('updated_at'))

    def test_get_recently_updated_returns_recent_beliefs(self):
        recent_belief = Belief(agent_id=self.agent.id, id="recent_belief_id")
        recent_belief.add_statement(self.statement)
        recent_belief.created_at = datetime.now()
        recent_belief.updated_at = datetime.now()
        
        old_statement = Statement(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=self.subject_notion,
            relation=Relation.from_string("causes"),
            complement_notion=self.complement_notion,
            probability=0.8,
            validity=0.9,
            id="old_stmt_id"
        )
        old_belief = Belief(agent_id=self.agent.id, id="old_belief_id")
        old_belief.add_statement(old_statement)
        old_belief.created_at = datetime.now() - timedelta(minutes=2)
        old_belief.updated_at = datetime.now() - timedelta(minutes=2)
        
        recent_belief_data = BeliefStorageMapper.to_dict(recent_belief)
        old_belief_data = BeliefStorageMapper.to_dict(old_belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)
        old_statement_data = StatementStorageMapper.to_dict(old_statement)

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements":
                if doc_id == self.statement.id:
                    return statement_data
                if doc_id == old_statement.id:
                    return old_statement_data
            return None

        self.storage_adapter.retrieve_by_attributes.return_value = [recent_belief_data, old_belief_data]
        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        one_minute_ago = datetime.now() - timedelta(minutes=1)
        results = self.repository.get_recently_updated(self.agent.id, one_minute_ago)
        
        expected_query = {'agent_id': "agent_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, "recent_belief_id")

    def test_get_recently_updated_returns_beliefs_updated_recently(self):
        belief = Belief(agent_id=self.agent.id, id="updated_belief_id")
        belief.add_statement(self.statement)
        belief.created_at = datetime.now() - timedelta(minutes=2)
        belief.updated_at = datetime.now()
        
        belief_data = BeliefStorageMapper.to_dict(belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]
        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        one_minute_ago = datetime.now() - timedelta(minutes=1)
        results = self.repository.get_recently_updated(self.agent.id, one_minute_ago)
        
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].id, "updated_belief_id")

    def test_get_recently_updated_excludes_old_beliefs(self):
        old_belief = Belief(agent_id=self.agent.id, id="old_belief_id")
        old_belief.add_statement(self.statement)
        old_belief.created_at = datetime.now() - timedelta(minutes=2)
        old_belief.updated_at = datetime.now() - timedelta(minutes=2)
        
        belief_data = BeliefStorageMapper.to_dict(old_belief)
        statement_data = StatementStorageMapper.to_dict(self.statement)

        def retrieve_by_id_side_effect(collection, doc_id):
            if collection == "statements" and doc_id == self.statement.id:
                return statement_data
            return None

        self.storage_adapter.retrieve_by_attributes.return_value = [belief_data]
        self.storage_adapter.retrieve_by_id.side_effect = retrieve_by_id_side_effect

        one_minute_ago = datetime.now() - timedelta(minutes=1)
        results = self.repository.get_recently_updated(self.agent.id, one_minute_ago)
        
        self.assertEqual(len(results), 0)

    def test_get_recently_updated_returns_empty_for_no_beliefs(self):
        self.storage_adapter.retrieve_by_attributes.return_value = []
        
        one_minute_ago = datetime.now() - timedelta(minutes=1)
        results = self.repository.get_recently_updated(self.agent.id, one_minute_ago)
        
        expected_query = {'agent_id': "agent_id"}
        self.storage_adapter.retrieve_by_attributes.assert_called_once_with("beliefs", expected_query)
        self.assertEqual(results, [])

    def tearDown(self):
        if self._config_override:
            self._config_override.__exit__(None, None, None)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import Mock, patch
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.user.user_service import UserService
from analogai.foundation.extensions.authentication.providers.auth0 import Auth0JWTService
from analogai.foundation.extensions.authentication.models import TokenPayload

class TestAuth0JWTService(unittest.TestCase):
    def setUp(self):
        self.user_service = Mock(spec=UserService)
        self.jwt_service = Auth0JWTService(self.user_service)
        self.user = User(id="test-user-123", name="Test User", email="test@example.com")

    def test_create_token_for_user(self):
        self.user_service.find.return_value = self.user
        with patch("analogai.foundation.extensions.authentication.providers.auth0.auth0_jwt_service.jwt.encode") as mock_encode:
            mock_encode.return_value = "mock.jwt.token"
            token = self.jwt_service.create_token_for_user(self.user.id)
            self.assertEqual(token, "mock.jwt.token")
            payload = mock_encode.call_args[0][0]
            self.assertEqual(payload["sub"], "analogai|test-user-123")
            self.assertEqual(payload["email"], "test@example.com")

    def test_create_token_for_user_missing_email(self):
        user_without_email = User(id="no-email", name="No Email", email=None)
        self.user_service.find.return_value = user_without_email
        with self.assertRaises(ValueError):
            self.jwt_service.create_token_for_user(user_without_email.id)

    def test_authenticate_user_success(self):
        self.user_service.find_by_email.return_value = self.user
        with patch.object(self.jwt_service, "extract_payload") as mock_extract:
            mock_extract.return_value = TokenPayload(token="t", email=self.user.email)
            result = self.jwt_service.authenticate_user("valid.token")
            self.assertEqual(result, self.user)

    def test_authenticate_user_no_user(self):
        self.user_service.find_by_email.return_value = None
        with patch.object(self.jwt_service, "extract_payload") as mock_extract:
            mock_extract.return_value = TokenPayload(token="t", email="missing@example.com")
            with self.assertRaises(ValueError):
                self.jwt_service.authenticate_user("valid.token")

    def test_authenticate_user_invalid_token(self):
        with patch.object(self.jwt_service, "extract_payload") as mock_extract:
            mock_extract.side_effect = ValueError("Invalid token format")
            with self.assertRaises(ValueError):
                self.jwt_service.authenticate_user("invalid.token")

    def test_extract_payload_success(self):
        mock_payload = {"email": "test@example.com", "sub": "auth0|user-123", "name": "Tester"}
        self.user_service.find_by_email.return_value = self.user
        with patch("analogai.foundation.extensions.authentication.providers.auth0.auth0_jwt_service.jwt.decode") as mock_decode:
            mock_decode.return_value = mock_payload
            payload = self.jwt_service.extract_payload("valid.token")
            self.assertEqual(payload.email, "test@example.com")
            self.assertEqual(payload.user_id, "test-user-123")
            self.assertEqual(payload.name, "Test User")
            self.assertEqual(payload.provider, "auth0")

    def test_extract_payload_missing_email(self):
        mock_payload = {"sub": "auth0|user-123"}
        with patch("analogai.foundation.extensions.authentication.providers.auth0.auth0_jwt_service.jwt.decode") as mock_decode:
            mock_decode.return_value = mock_payload
            with self.assertRaises(ValueError):
                self.jwt_service.extract_payload("valid.token")

if __name__ == "__main__":
    unittest.main()




from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from importlib import resources
from typing import Any, Dict

from analogai.foundation.common.utils.config_validator import validate_settings


_DEFAULT_SCHEMA_PATH = resources.files("analogai.foundation.config") / "schema.yaml"


@dataclass
class ConfigBase:
    settings: object
    schema_path: str | Path = _DEFAULT_SCHEMA_PATH
    _overrides: Dict[str, Any] = field(default_factory=dict, init=False, repr=False)

    def __post_init__(self) -> None:
        validate_settings(self.settings, self.schema_path)

    def get(self, key: str, default: Any = None) -> Any:
        if key in self._overrides:
            return self._overrides[key]
        return self.settings.get(key, default)

    @contextmanager
    def override(self, **kwargs: Any):
        old: Dict[str, Any] = {}
        for k, v in kwargs.items():
            old[k] = self._overrides.get(k, None)
            self._overrides[k] = v
        try:
            yield
        finally:
            for k in kwargs:
                if k in old and old[k] is not None:
                    self._overrides[k] = old[k]
                else:
                    self._overrides.pop(k, None)



from analogai.foundation.extensions.authentication.authentication_factory import AuthenticationFactory
from analogai.foundation.extensions.authentication.interfaces.authentication_factory_interface import (
    AuthenticationDependency,
    TokenExtractionDependency,
    ApiTokenAuthenticationDependency,
)

def build_authentication_dependency() -> AuthenticationDependency:
    factory = AuthenticationFactory()
    return factory.create_authentication_dependency()

def build_token_extraction_dependency() -> TokenExtractionDependency:
    factory = AuthenticationFactory()
    return factory.create_token_extraction_dependency()

def build_api_token_authentication_dependency() -> ApiTokenAuthenticationDependency:
    factory = AuthenticationFactory()
    return factory.create_api_token_authentication_dependency()




import unittest
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.common.enums.deontic_modality import DeonticModality

class TestModifier(unittest.TestCase):
    
    def test_create_empty_context(self):
        context = Modifier()
        
        self.assertIsNone(context.location)
        self.assertIsNone(context.deontic_modality)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_create_context_with_location(self):
        context = Modifier(location="New York")
        
        self.assertEqual(context.location, "New York")
        self.assertIsNone(context.deontic_modality)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_create_context_with_deontic_modality(self):
        context = Modifier(deontic_modality=DeonticModality.OBLIGATORY)
        
        self.assertIsNone(context.location)
        self.assertEqual(context.deontic_modality, DeonticModality.OBLIGATORY)
        self.assertIsNone(context.from_time)
        self.assertIsNone(context.to_time)
    
    def test_create_context_with_time(self):
        from_time = Time(year=2024, month=1, day=1)
        to_time = Time(year=2024, month=12, day=31)
        context = Modifier(from_time=from_time, to_time=to_time)
        
        self.assertIsNone(context.location)
        self.assertIsNone(context.deontic_modality)
        self.assertEqual(context.from_time, from_time)
        self.assertEqual(context.to_time, to_time)
    
    def test_create_context_with_all_fields(self):
        from_time = Time(year=2024, month=1, day=1)
        to_time = Time(year=2024, month=12, day=31)
        context = Modifier(
            location="London",
            deontic_modality=DeonticModality.PERMITTED,
            from_time=from_time,
            to_time=to_time
        )
        
        self.assertEqual(context.location, "London")
        self.assertEqual(context.deontic_modality, DeonticModality.PERMITTED)
        self.assertEqual(context.from_time, from_time)
        self.assertEqual(context.to_time, to_time)

if __name__ == '__main__':
    unittest.main()



from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.settings import config
from analogai.foundation.modules.notion.notion_repository import NotionRepository
from analogai.foundation.common.utils.attribute_key_resolvers import AttributeKeyResolverChain
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.core.observables.notion_observable_service import NotionObservableService
from typeguard import typechecked
from typing import Optional, List


def _environment() -> str:
    value = config.get("ENVIRONMENT")
    return str(value).strip().lower()


def _use_vector_search() -> bool:
    if _environment() in {"test", "testing"}:
        return False
    return bool(config.get("vector.enabled", True))


def _similarity_cutoff() -> float:
    value = config.get("vector.cutoff", 0.4)
    try:
        return float(value)
    except (TypeError, ValueError):
        return 0.4


class NotionService(NotionObservableService):
    def __init__(self, notion_repository: NotionRepository):
        NotionObservableService.__init__(self)
        self.notion_repository = notion_repository

    @typechecked
    def create(self, user_id: str, agent_id: str, expression: NotionExpression) -> Notion:
        notion = Notion(
            user_id=user_id,
            agent_id=agent_id,
            caption=expression.caption,
            attributes=expression.attributes,
        )
        created_notion = self.notion_repository.create(notion)
        self.notify_notion_created(created_notion)
        return created_notion
    
    @typechecked
    def find_or_create(
        self,
        user_id: str,
        agent_id: str,
        expression: NotionExpression,
        similarity_threshold: Optional[float] = None,
    ) -> Notion:
        if similarity_threshold is None:
            existing = self.notion_repository.find(
                agent_id,
                expression,
                None,
                similarity_threshold=1.0,
            )
            if (
                existing
                and existing.caption.strip().lower() == expression.caption.strip().lower()
            ):
                return existing
        else:
            existing = self.notion_repository.find(
                agent_id,
                expression,
                None,
                similarity_threshold=similarity_threshold,
            )
            if existing:
                return existing

        return self.create(user_id, agent_id, expression)

    @typechecked
    def upsert_from_expression(
        self,
        user_id: str,
        agent_id: str,
        expression: NotionExpression,
        similarity_threshold: Optional[float] = None,
    ) -> Notion:
        existing = self.find(agent_id, expression, similarity_threshold=similarity_threshold)
        merged_attributes = self._merge_attributes(
            existing.attributes if existing else None,
            expression.attributes,
        )
        notion = Notion(
            user_id=user_id,
            agent_id=agent_id,
            caption=expression.caption,
            id=existing.id if existing else None,
            attributes=merged_attributes,
        )
        created = self.notion_repository.create(notion)
        if existing is None:
            self.notify_notion_created(created)
        return created
        
    @typechecked
    def find(
        self,
        agent_id: str,
        notion_expression: NotionExpression,
        similarity_threshold: Optional[float] = None,
    ) -> Optional[Notion]:
        if similarity_threshold is None and _use_vector_search():
            similarity_threshold = _similarity_cutoff()
        return self.notion_repository.find(
            agent_id,
            notion_expression,
            None,
            similarity_threshold=similarity_threshold,
        )

    @typechecked
    def get_all_for_agent(self, agent_id: str) -> List[Notion]:
        return self.notion_repository.find_all_by_agent(agent_id)

    @staticmethod
    def _merge_attributes(
        existing: Optional[List[object]],
        incoming: Optional[List[object]],
    ) -> Optional[List[object]]:
        if not existing and not incoming:
            return None
        merged: List[object] = []
        seen = set()
        key_resolver = AttributeKeyResolverChain()
        for attribute_group in (existing or [], incoming or []):
            for attribute in attribute_group:
                attribute_key = key_resolver.resolve(attribute)
                if attribute_key in seen:
                    continue
                seen.add(attribute_key)
                merged.append(attribute)
        return merged


import uuid
from typing import Optional, List

from analogai.foundation.core.value_objects.attribute import Attribute
from analogai.foundation.common.utils.entity_formatters import (
    format_notion_str,
    format_notion_repr,
)


class Notion:
    def __init__(
        self,
        user_id: str,
        agent_id: str,
        caption: str,
        id: Optional[str] = None,
        attributes: Optional[List[Attribute]] = None,
    ):
        self.id = id if id else str(uuid.uuid4())
        self.user_id = user_id
        self.agent_id = agent_id
        self.caption = caption
        self.attributes = attributes

    def __str__(self) -> str:
        return format_notion_str(self)

    def __repr__(self) -> str:
        return format_notion_repr(self)




from typing import List
import math

from analogai.foundation.common.utils.similarity.similarity_service import SimilarityService
from analogai.foundation.infrastructure.services.embedding_service import EmbeddingService


class EmbeddingSimilarityService(SimilarityService):
    def __init__(self, embedding_service: EmbeddingService | None = None):
        self._embedding_service = embedding_service or EmbeddingService()

    def similarity(self, text_a: str, text_b: str) -> float:
        vector_a = self._embedding_service.encode(text_a)
        vector_b = self._embedding_service.encode(text_b)
        return self._cosine_similarity(vector_a, vector_b)

    @staticmethod
    def _cosine_similarity(vector_a: List[float], vector_b: List[float]) -> float:
        if not vector_a or not vector_b or len(vector_a) != len(vector_b):
            return 0.0
        dot = sum(a * b for a, b in zip(vector_a, vector_b))
        norm_a = math.sqrt(sum(a * a for a in vector_a))
        norm_b = math.sqrt(sum(b * b for b in vector_b))
        if norm_a == 0 or norm_b == 0:
            return 0.0
        return dot / (norm_a * norm_b)



import importlib
import os
import unittest
from unittest.mock import patch

from analogai.foundation.common.design_patterns.singleton import Singleton
from analogai.foundation.common.utils.env_utils import load_envs_in_order
from analogai.foundation.settings import config


class TestEmbeddingServiceFunctionality(unittest.TestCase):
    def setUp(self):
        self._env_backup = dict(os.environ)
        load_envs_in_order()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self._env_backup)
        from analogai.foundation.infrastructure.services import embedding_service

        Singleton._instances.pop(embedding_service.EmbeddingService, None)

    def _reload_embedding_modules(self):
        from analogai.foundation.infrastructure.embeddings import provider_factory
        from analogai.foundation.infrastructure.services import embedding_service
        from analogai.foundation import settings

        importlib.reload(settings)
        importlib.reload(provider_factory)
        importlib.reload(embedding_service)
        return settings.config, provider_factory, embedding_service

    def test_embedding_provider_encode(self):
        _config, _provider_factory, embedding_service = self._reload_embedding_modules()
        environment = _config.get("ENVIRONMENT")
        use_vector_search = environment not in {"test", "testing"} and bool(
            _config.get("vector.enabled", True)
        )
        if not use_vector_search:
            self.skipTest("USE_VECTOR_SEARCH is false; skipping embedding tests")

        if str(_config.get("embed.type", "local")).strip().lower() == "local":
            try:
                import sentence_transformers  # noqa: F401
            except Exception as exc:
                self.fail(f"sentence_transformers is required for local embeddings: {exc}")

            os.environ.setdefault("EMBEDDING_MODEL_NAME", "all-MiniLM-L6-v2")
        elif str(_config.get("embed.type", "local")).strip().lower() == "api":
            provider = str(_config.get("embed.provider", "")).strip().lower()
            if provider not in {"openai", "azure_openai", "azure_foundry", "azure"}:
                self.fail(
                    "EMBEDDING_PROVIDER must be one of: openai, azure_openai, azure_foundry, azure"
                )
            if not _config.get("embed.model", ""):
                self.fail("EMBEDDING_MODEL_NAME must be set for API embeddings")
            if provider in {"openai"} and not _config.get("OPENAI_API_KEY", ""):
                self.fail("OPENAI_API_KEY is required for OpenAI embeddings")
            if provider in {"azure", "azure_openai"} and (
                not _config.get("AZURE_OPENAI_ENDPOINT", "")
                or not _config.get("AZURE_OPENAI_API_KEY", "")
            ):
                self.fail("AZURE_OPENAI_ENDPOINT and AZURE_OPENAI_API_KEY are required")
            if provider in {"azure_foundry"} and (
                not _config.get("AZURE_AI_FOUNDRY_ENDPOINT", "")
                or not _config.get("AZURE_AI_FOUNDRY_API_KEY", "")
            ):
                self.fail("AZURE_AI_FOUNDRY_ENDPOINT and AZURE_AI_FOUNDRY_API_KEY are required")
        else:
            self.fail("EMBEDDING_TYPE must be 'local' or 'api'")

        Singleton._instances.pop(embedding_service.EmbeddingService, None)
        service = embedding_service.EmbeddingService()
        vector = service.encode("hello")
        self.assertTrue(len(vector) > 0)

    def test_embedding_provider_batch_encode(self):
        _config, _provider_factory, embedding_service = self._reload_embedding_modules()
        environment = _config.get("ENVIRONMENT")
        use_vector_search = environment not in {"test", "testing"} and bool(
            _config.get("vector.enabled", True)
        )
        if not use_vector_search:
            self.skipTest("USE_VECTOR_SEARCH is false; skipping embedding tests")

        Singleton._instances.pop(embedding_service.EmbeddingService, None)
        service = embedding_service.EmbeddingService()
        batch = service.encode_batch(["a", "b"])
        self.assertEqual(len(batch), 2)
        self.assertTrue(len(batch[0]) > 0)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from .observable_service import ObservableService


class UserObserver(ABC):
    @abstractmethod
    def on_user_created(self, user):
        pass


class UserObservableService(ObservableService):
    def notify_user_created(self, user) -> None:
        for observer in self._observers:
            try:
                observer.on_user_created(user)
            except (AttributeError, TypeError):
                pass